__all__ = ["YarNumBaseError", "NonIntegerInputError", "InvalidFormError"]


class YarNumBaseError(Exception):
    pass


class NonIntegerInputError(YarNumBaseError):
    def __init__(self, v: object) -> None:
        super().__init__(f"Provided value is not a valid integer: {str(v)}")


class InvalidFormError(YarNumBaseError):
    def __init__(self, v: object) -> None:
        super().__init__(f"Provided inflection form is not valid: {str(v)}")


class RangeError(YarNumBaseError):
    def __init__(self, n, max_len: int) -> None:
        super().__init__(
            f"Provided number is too large with {len(n)} out of {max_len} digits."
        )


class InternalDataError(YarNumBaseError):
    def __init__(self) -> None:
        super().__init__(f"Numeral in the requested form not found in internal data")



import pytest
from .ordinals import inflect
from .errors import RangeError

# fmt: off
simple_map = {
    "nom": ("case", "nominative"),
    "gen": ("case", "genitive"),
    "dat": ("case", "dative"),
    "acc": ("case", "accusative"),
    "inst": ("case", "instrumental"),
    "loc": ("case", "locative"),
    "voc": ("case", "vocative"),
    "m": ("gender", "masculine"),
    "f": ("gender", "feminine"),
    "n": ("gender", "neuter"),
    "sg": ("number", "singular"),
    "pl": ("number", "plural"),
    "card": ("type", "cardinal"),
    "ord": ("type", "ordinal"),
    "inan": ("animacy", "inanimate"),
    "anim": ("animacy", "animate"),
}


def fm(s):
    return {simple_map[g][0]: simple_map[g][1] for g in s.split(";")}


def test_0():
    assert inflect("0", fm("m")) == "нульови́й"
    assert inflect("0", fm("f")) == "нульова́"
    assert inflect("0", fm("n")) == "нульове́"
    assert inflect("0", fm("gen")) == "нульово́го"
    assert inflect("0", fm("gen;f")) == "нульово́ї"
    assert inflect("0", fm("acc;f;anim")) == "нульову́"
    assert inflect("0", fm("acc;f;inan")) == "нульову́"
    assert inflect("0", fm("loc;f")) == "нульові́й"


def test_lpad_0():
    assert inflect("00", {}) == "нульови́й"
    assert inflect("000000", {}) == "нульови́й"
    assert inflect("021", fm("m")) == "два́дцять пе́рший"
    assert inflect("0000000021", fm("m")) == "два́дцять пе́рший"


def test_21():
    assert inflect("21", fm("m")) == "два́дцять пе́рший"
    assert inflect("21", fm("f")) == "два́дцять пе́рша"
    assert inflect("21", fm("n")) == "два́дцять пе́рше"
    assert inflect("21", fm("gen")) == "два́дцять пе́ршого"
    assert inflect("21", fm("gen;f")) == "два́дцять пе́ршої"
    assert inflect("21", fm("acc;f;anim")) == "два́дцять пе́ршу"
    assert inflect("21", fm("acc;f;inan")) == "два́дцять пе́ршу"
    assert inflect("21", fm("loc;f")) == "два́дцять пе́ршій"


def test_713():
    assert inflect("713", fm("m")) == "сімсо́т трина́дцятий"
    assert inflect("713", fm("f")) == "сімсо́т трина́дцята"
    assert inflect("713", fm("n")) == "сімсо́т трина́дцяте"
    assert inflect("713", fm("gen")) == "сімсо́т трина́дцятого"
    assert inflect("713", fm("gen;f")) == "сімсо́т трина́дцятої"
    assert inflect("713", fm("acc;f;anim")) == "сімсо́т трина́дцяту"
    assert inflect("713", fm("acc;f;inan")) == "сімсо́т трина́дцяту"
    assert inflect("713", fm("loc;f")) == "сімсо́т трина́дцятій"


def test_1000():
    assert inflect("1000", fm("m")) == "ти́сячний"
    assert inflect("1000", fm("f")) == "ти́сячна"
    assert inflect("1000", fm("n")) == "ти́сячне"
    assert inflect("1000", fm("gen;m")) == "ти́сячного"
    assert inflect("1000", fm("gen;f")) == "ти́сячної"
    assert inflect("1000", fm("dat")) == "ти́сячному"
    assert inflect("1000", fm("acc;m;inan")) == "ти́сячний"
    assert inflect("1000", fm("acc;m;anim")) == "ти́сячного"
    assert inflect("1000", fm("acc;f;inan")) == "ти́сячну"
    assert inflect("1000", fm("acc;f;anim")) == "ти́сячну"
    assert inflect("1000", fm("acc;n;inan")) == "ти́сячне"
    assert inflect("1000", fm("acc;n;anim")) == "ти́сячне"
    assert inflect("1000", fm("acc;pl;inan")) == "ти́сячні"
    assert inflect("1000", fm("acc;pl;anim")) == "ти́сячних"
    assert inflect("1000", fm("loc;f")) == "ти́сячній"
    assert inflect("1000", fm("loc")) == "ти́сячному"
    assert inflect("1000", fm("inst")) == "ти́сячним"


def test_4486():
    assert inflect("4486", fm("m")) == "чоти́ри ти́сячі чоти́риста вісімдеся́т шо́стий"
    assert inflect("4486", fm("f")) == "чоти́ри ти́сячі чоти́риста вісімдеся́т шо́ста"
    assert inflect("4486", fm("n")) == "чоти́ри ти́сячі чоти́риста вісімдеся́т шо́сте"
    assert inflect("4486", fm("pl")) == "чоти́ри ти́сячі чоти́риста вісімдеся́т шо́сті"

    assert inflect("4486", fm("gen;m")) == "чоти́ри ти́сячі чоти́риста вісімдеся́т шо́стого"
    assert inflect("4486", fm("gen;f")) == "чоти́ри ти́сячі чоти́риста вісімдеся́т шо́стої"
    assert inflect("4486", fm("gen;n")) == "чоти́ри ти́сячі чоти́риста вісімдеся́т шо́стого"
    assert inflect("4486", fm("gen;pl")) == "чоти́ри ти́сячі чоти́риста вісімдеся́т шо́стих"

    assert inflect("4486", fm("dat;m")) == "чоти́ри ти́сячі чоти́риста вісімдеся́т шо́стому"
    assert inflect("4486", fm("dat;f")) == "чоти́ри ти́сячі чоти́риста вісімдеся́т шо́стій"
    assert inflect("4486", fm("dat;n")) == "чоти́ри ти́сячі чоти́риста вісімдеся́т шо́стому"
    assert inflect("4486", fm("dat;pl")) == "чоти́ри ти́сячі чоти́риста вісімдеся́т шо́стим"

    assert inflect("4486", fm("acc;m;inan")) == "чоти́ри ти́сячі чоти́риста вісімдеся́т шо́стий"
    assert inflect("4486", fm("acc;f;inan")) == "чоти́ри ти́сячі чоти́риста вісімдеся́т шо́сту"
    assert inflect("4486", fm("acc;n;inan")) == "чоти́ри ти́сячі чоти́риста вісімдеся́т шо́сте"
    assert inflect("4486", fm("acc;pl;inan")) == "чоти́ри ти́сячі чоти́риста вісімдеся́т шо́сті"

    assert inflect("4486", fm("acc;m;anim")) == "чоти́ри ти́сячі чоти́риста вісімдеся́т шо́стого"
    assert inflect("4486", fm("acc;f;anim")) == "чоти́ри ти́сячі чоти́риста вісімдеся́т шо́сту"
    assert inflect("4486", fm("acc;n;anim")) == "чоти́ри ти́сячі чоти́риста вісімдеся́т шо́сте"
    assert inflect("4486", fm("acc;pl;anim")) == "чоти́ри ти́сячі чоти́риста вісімдеся́т шо́стих"

    assert inflect("4486", fm("inst;m")) == "чоти́ри ти́сячі чоти́риста вісімдеся́т шо́стим"
    assert inflect("4486", fm("inst;f")) == "чоти́ри ти́сячі чоти́риста вісімдеся́т шо́стою"
    assert inflect("4486", fm("inst;n")) == "чоти́ри ти́сячі чоти́риста вісімдеся́т шо́стим"
    assert inflect("4486", fm("inst;pl")) == "чоти́ри ти́сячі чоти́риста вісімдеся́т шо́стими"

    assert inflect("4486", fm("loc;m")) == "чоти́ри ти́сячі чоти́риста вісімдеся́т шо́стому"
    assert inflect("4486", fm("loc;f")) == "чоти́ри ти́сячі чоти́риста вісімдеся́т шо́стій"
    assert inflect("4486", fm("loc;n")) == "чоти́ри ти́сячі чоти́риста вісімдеся́т шо́стому"
    assert inflect("4486", fm("loc;pl")) == "чоти́ри ти́сячі чоти́риста вісімдеся́т шо́стих"

    assert inflect("4486", fm("voc;m")) == "чоти́ри ти́сячі чоти́риста вісімдеся́т шо́стий"
    assert inflect("4486", fm("voc;f")) == "чоти́ри ти́сячі чоти́риста вісімдеся́т шо́ста"
    assert inflect("4486", fm("voc;n")) == "чоти́ри ти́сячі чоти́риста вісімдеся́т шо́сте"
    assert inflect("4486", fm("voc;pl")) == "чоти́ри ти́сячі чоти́риста вісімдеся́т шо́сті"


def test_525000():
    assert inflect("525000", fm("m")) == "пʼятисотдвадцятипʼятити́сячний"
    assert inflect("525000", fm("f")) == "пʼятисотдвадцятипʼятити́сячна"
    assert inflect("525000", fm("n")) == "пʼятисотдвадцятипʼятити́сячне"
    assert inflect("525000", fm("pl")) == "пʼятисотдвадцятипʼятити́сячні"

    assert inflect("525000", fm("gen;m")) == "пʼятисотдвадцятипʼятити́сячного"
    assert inflect("525000", fm("gen;f")) == "пʼятисотдвадцятипʼятити́сячної"
    assert inflect("525000", fm("gen;n")) == "пʼятисотдвадцятипʼятити́сячного"
    assert inflect("525000", fm("gen;pl")) == "пʼятисотдвадцятипʼятити́сячних"

    assert inflect("525000", fm("dat;m")) == "пʼятисотдвадцятипʼятити́сячному"
    assert inflect("525000", fm("dat;f")) == "пʼятисотдвадцятипʼятити́сячній"
    assert inflect("525000", fm("dat;n")) == "пʼятисотдвадцятипʼятити́сячному"
    assert inflect("525000", fm("dat;pl")) == "пʼятисотдвадцятипʼятити́сячним"

    assert inflect("525000", fm("acc;m;inan")) == "пʼятисотдвадцятипʼятити́сячний"
    assert inflect("525000", fm("acc;f;inan")) == "пʼятисотдвадцятипʼятити́сячну"
    assert inflect("525000", fm("acc;n;inan")) == "пʼятисотдвадцятипʼятити́сячне"
    assert inflect("525000", fm("acc;pl;inan")) == "пʼятисотдвадцятипʼятити́сячні"

    assert inflect("525000", fm("acc;m;anim")) == "пʼятисотдвадцятипʼятити́сячного"
    assert inflect("525000", fm("acc;f;anim")) == "пʼятисотдвадцятипʼятити́сячну"
    assert inflect("525000", fm("acc;n;anim")) == "пʼятисотдвадцятипʼятити́сячне"
    assert inflect("525000", fm("acc;pl;anim")) == "пʼятисотдвадцятипʼятити́сячних"

    assert inflect("525000", fm("inst;m")) == "пʼятисотдвадцятипʼятити́сячним"
    assert inflect("525000", fm("inst;f")) == "пʼятисотдвадцятипʼятити́сячною"
    assert inflect("525000", fm("inst;n")) == "пʼятисотдвадцятипʼятити́сячним"
    assert inflect("525000", fm("inst;pl")) == "пʼятисотдвадцятипʼятити́сячними"

    assert inflect("525000", fm("loc;m")) == "пʼятисотдвадцятипʼятити́сячному"
    assert inflect("525000", fm("loc;f")) == "пʼятисотдвадцятипʼятити́сячній"
    assert inflect("525000", fm("loc;n")) == "пʼятисотдвадцятипʼятити́сячному"
    assert inflect("525000", fm("loc;pl")) == "пʼятисотдвадцятипʼятити́сячних"

    assert inflect("525000", fm("voc;m")) == "пʼятисотдвадцятипʼятити́сячний"
    assert inflect("525000", fm("voc;f")) == "пʼятисотдвадцятипʼятити́сячна"
    assert inflect("525000", fm("voc;n")) == "пʼятисотдвадцятипʼятити́сячне"
    assert inflect("525000", fm("voc;pl")) == "пʼятисотдвадцятипʼятити́сячні"


def test_gender_with_plural():
    assert inflect("525000", fm("m;pl")) == "пʼятисотдвадцятипʼятити́сячні"
    assert inflect("525000", fm("gen;m;pl")) == "пʼятисотдвадцятипʼятити́сячних"
    assert inflect("525000", fm("dat;f;pl")) == "пʼятисотдвадцятипʼятити́сячним"
    assert inflect("525000", fm("loc;f;pl;anim")) == "пʼятисотдвадцятипʼятити́сячних"


def test_pravopys_p38():
    """§ 38 Складні числівники"""

    assert inflect("900", fm("m"), False) == "девʼятисо́тий"
    assert inflect("300", fm("m"), False) == "трьохсо́тий"
    assert inflect("2000", fm("m"), False) == "двохти́сячний"
    assert inflect("10000", fm("m"), False) == "десятити́сячний"
    assert inflect("4000000", fm("m"), False) == "чотирьохмільйо́нний"
    assert inflect("7000000000", fm("m"), False) == "семимілья́рдний"
    assert inflect("3000000000", fm("m"), False) == "трьохмілья́рдний"
    assert inflect("65000000", fm("m"), False) == "шістдесятипʼятимільйо́нний"
    assert inflect("1003000", fm("m"), False) == "мільйо́н трьохти́сячний"
    assert inflect("1004000000", fm("m"), False) == "мілья́рд чотирьохмільйо́нний"
    assert inflect("1988", fm("m"), False) == "ти́сяча девʼятсо́т вісімдеся́т во́сьмий"


def test_pravopys_p38_error():
    # There is an error in Правопис which led to incorrect example being printed
    assert inflect("530000", fm("m"), False) != "пʼятсоттридцятити́сячний"
    assert inflect("50000000", fm("m"), False) != "пʼятдесятимільйо́нний"


def test_pravopys_p38_corrected():
    assert inflect("530000", fm("m"), False) == "пʼятисоттридцятити́сячний"
    assert inflect("50000000", fm("m"), False) == "пʼятидесятимільйо́нний"


def test_pravopys_p106():
    """§ 106 Відмінювання порядкових числівників"""

    assert inflect("1", fm("m"), False) == "пе́рший"
    assert inflect("1", fm("f"), False) == "пе́рша"
    assert inflect("1", fm("n"), False) == "пе́рше"
    assert inflect("2", fm("m"), False) == "дру́гий"
    assert inflect("4", fm("m"), False) == "четве́ртий"
    assert inflect("5", fm("m"), False) == "пʼя́тий"
    assert inflect("6", fm("m"), False) == "шо́стий"
    assert inflect("7", fm("m"), False) == "сьо́мий"
    assert inflect("8", fm("m"), False) == "во́сьмий"
    assert inflect("9", fm("m"), False) == "девʼя́тий"
    assert inflect("10", fm("m"), False) == "деся́тий"
    assert inflect("11", fm("m"), False) == "одина́дцятий"
    assert inflect("12", fm("m"), False) == "двана́дцятий"
    assert inflect("20", fm("m"), False) == "двадця́тий"
    assert inflect("30", fm("m"), False) == "тридця́тий"
    assert inflect("40", fm("m"), False) == "сороко́вий"
    assert inflect("50", fm("m"), False) == "пʼятдеся́тий"
    assert inflect("60", fm("m"), False) == "шістдеся́тий"
    assert inflect("70", fm("m"), False) == "сімдеся́тий"
    assert inflect("80", fm("m"), False) == "вісімдеся́тий"
    assert inflect("90", fm("m"), False) == "девʼяно́стий"
    assert inflect("100", fm("m"), False) == "со́тий"
    assert inflect("200", fm("m"), False) == "двохсо́тий"
    assert inflect("300", fm("m"), False) == "трьохсо́тий"
    assert inflect("400", fm("m"), False) == "чотирьохсо́тий"
    assert inflect("500", fm("m"), False) == "пʼятисо́тий"
    assert inflect("1000", fm("m"), False) == "ти́сячний"
    assert inflect("2000", fm("m"), False) == "двохти́сячний"
    assert inflect("3000", fm("m"), False) == "трьохти́сячний"
    assert inflect("4000", fm("m"), False) == "чотирьохти́сячний"
    assert inflect("5000", fm("m"), False) == "пʼятити́сячний"
    assert inflect("1000000", fm("m"), False) == "мільйо́нний"
    assert inflect("2000000", fm("m"), False) == "двохмільйо́нний"
    assert inflect("3000000", fm("m"), False) == "трьохмільйо́нний"
    assert inflect("4000000", fm("m"), False) == "чотирьохмільйо́нний"
    assert inflect("5000000", fm("m"), False) == "пʼятимільйо́нний"
    assert inflect("3", fm("m"), False) == "тре́тій"
    assert inflect("3", fm("f"), False) == "тре́тя"
    assert inflect("3", fm("n"), False) == "тре́тє"
    assert inflect("88", fm("m"), False) == "вісімдеся́т во́сьмий"
    assert inflect("88", fm("gen;m"), False) == "вісімдеся́т во́сьмого"
    assert inflect("1991", fm("gen;m"), False) == "ти́сяча девʼятсо́т девʼяно́сто пе́ршого"
    assert inflect("1789", fm("loc;m"), False) == "ти́сяча сімсо́т вісімдеся́т девʼя́тому"
    assert inflect("2018", fm("dat;m"), False) == "дві́ ти́сячі вісімна́дцятому"


def test_range_fallback():
    assert inflect("999999999999999999999999999", fm("m")) == "девʼятсо́т девʼяно́сто де́вʼять септильйо́нів девʼятсо́т девʼяно́сто де́вʼять секстильйо́нів девʼятсо́т девʼяно́сто де́вʼять квінтильйо́нів девʼятсо́т девʼяно́сто де́вʼять квадрильйо́нів девʼятсо́т девʼяно́сто де́вʼять трильйо́нів девʼятсо́т девʼяно́сто де́вʼять мілья́рдів девʼятсо́т девʼяно́сто де́вʼять мільйо́нів девʼятсо́т девʼяно́сто де́вʼять ти́сяч девʼятсо́т девʼяно́сто девʼя́тий"
    assert inflect("1000000000000000000000000000", fm("m")) == "оди́н ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль нульови́й"
    assert inflect("1000000000000000000000000001", fm("m")) == "оди́н ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль пе́рший"
    assert inflect("1234567890123456789012345678", fm("m")) == "оди́н два́ три́ чоти́ри пʼя́ть ші́сть сі́м ві́сім де́вʼять ну́ль оди́н два́ три́ чоти́ри пʼя́ть ші́сть сі́м ві́сім де́вʼять ну́ль оди́н два́ три́ чоти́ри пʼя́ть ші́сть сі́м во́сьмий"
    assert inflect("999999999999999999999999999", fm("f")) == "девʼятсо́т девʼяно́сто де́вʼять септильйо́нів девʼятсо́т девʼяно́сто де́вʼять секстильйо́нів девʼятсо́т девʼяно́сто де́вʼять квінтильйо́нів девʼятсо́т девʼяно́сто де́вʼять квадрильйо́нів девʼятсо́т девʼяно́сто де́вʼять трильйо́нів девʼятсо́т девʼяно́сто де́вʼять мілья́рдів девʼятсо́т девʼяно́сто де́вʼять мільйо́нів девʼятсо́т девʼяно́сто де́вʼять ти́сяч девʼятсо́т девʼяно́сто девʼя́та"
    assert inflect("1000000000000000000000000000", fm("f")) == "оди́н ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль нульова́"
    assert inflect("1000000000000000000000000001", fm("f")) == "оди́н ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль пе́рша"
    assert inflect("1234567890123456789012345678", fm("f")) == "оди́н два́ три́ чоти́ри пʼя́ть ші́сть сі́м ві́сім де́вʼять ну́ль оди́н два́ три́ чоти́ри пʼя́ть ші́сть сі́м ві́сім де́вʼять ну́ль оди́н два́ три́ чоти́ри пʼя́ть ші́сть сі́м во́сьма"
    assert inflect("1000000000000000000000000000", fm("gen;f")) == "оди́н ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль нульово́ї"
    assert inflect("1000000000000000000000000001", fm("gen;f")) == "оди́н ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль пе́ршої"


def test_strict_range():
    inflect("999999999999999999999999999", fm("m"), strict_range=True)
    with pytest.raises(RangeError):
        assert inflect("1000000000000000000000000000", fm("m"), strict_range=True)


def test_10000000000000000000000000():
    assert inflect("100000000000000000000000000", fm("m")) == "стасептильйо́нний"


def test_insert_ones():
    assert inflect("100", fm("m"), insert_ones=True) == "со́тий"
    assert inflect("101", fm("m"), insert_ones=True) == "сто́ пе́рший"
    assert inflect("1000", fm("m"), insert_ones=True) == "ти́сячний"
    assert inflect("1001", fm("m"), insert_ones=True) == "одна́ ти́сяча пе́рший"
    assert inflect("1000000", fm("m"), insert_ones=True) == "мільйо́нний"
    assert inflect("1000001", fm("m"), insert_ones=True) == "оди́н мільйо́н пе́рший"
    assert inflect("100", fm("m"), insert_ones=True) == "со́тий"
    assert inflect("101", fm("f"), insert_ones=True) == "сто́ пе́рша"
    assert inflect("1000", fm("f"), insert_ones=True) == "ти́сячна"
    assert inflect("1001", fm("f"), insert_ones=True) == "одна́ ти́сяча пе́рша"
    assert inflect("1000000", fm("f"), insert_ones=True) == "мільйо́нна"
    assert inflect("1000001", fm("f"), insert_ones=True) == "оди́н мільйо́н пе́рша"


# fmt: on



from .grapheme import remove_stress


def test_remove_stress():
    assert remove_stress("") == ""
    assert remove_stress("оди́н") == "один"
    assert remove_stress("оди́н мільйо́н") == "один мільйон"
    assert remove_stress("о́о́о́о́") == "оооо"
    assert remove_stress("один") == "один"



# (c) Copyright by Yar.
#
# The basic API definitions for the numerals module.

from . import (
    cardinals,
    decimals,
    fractionals,
    grammar,
    errors,
    grapheme,
    ordinals,
    validator,
)

__all__ = [
    "inflect_cardinal",
    "inflect_ordinal",
    "inflect_decimal",
    "inflect_fractional",
    "grammar",
    "errors",
]


def inflect_cardinal(
    whole: str,
    form: grammar.InflectionForm,
    stress: bool = False,
    insert_ones: bool = False,
    strict_range: bool = False,
) -> str:
    """
    Returns a spelling of the provided non-negative integer number as a cardinal
    numeral. The spelling is inflected for a requested form. The input is validated and
    exception of type descending from :any:`errors.YarNumBaseError` is thrown if
    validation fails. The range of numbers that can be correctly inflected is [0...1e27);
    numbers outside this range are not processed correctly. See `strict_range` for more
    details.

    :param whole: The string of digits in base 10 to be spelled out.
    :param form: The grammatical form of the generated text.
    :param stress: Whether the output text should contain stresses.
    :param insert_ones: Whether the output should insert "ones" in thousands and millions.
    :param strict_range: Whether to throw an exception when the number is too big to
    correctly process.
    :return: The spellout of the provided number as a cardinal numeral inflected for a given form.

    Basic usage
        >>> inflect_cardinal("43", {"case": "nominative"})
        'сорок три'
        >>> inflect_cardinal("43", {"case": "genitive"})
        'сорока трьох'

    Using `stress` parameter
        >>> inflect_cardinal("75", {}, stress=False)
        'сімдесят пʼять'
        >>> inflect_cardinal("75", {}, stress=True)
        'сімдеся́т пʼя́ть'

    Using `insert_ones` parameter
        >>> inflect_cardinal("1000", {}, insert_ones=False)
        'тисяча'
        >>> inflect_cardinal("1000", {}, insert_ones=True)
        'одна тисяча'

    Using `strict_range` parameter
        >>> inflect_cardinal("100000000000000000000000000", {}, strict_range=False)
        'сто септильйонів'
        >>> inflect_cardinal("1000000000000000000000000000", {}, strict_range=False)
        'один нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль'
        >>> inflect_cardinal("1000000000000000000000000000", {}, strict_range=True)
        Traceback (most recent call last):
            ...
        yar_numerals.errors.RangeError: Provided number is too large with 28 out of 27 digits.
    """
    validator.validate_int(whole)
    validator.validate_form(form)
    text = cardinals.inflect(whole, form, insert_ones, strict_range)
    if not stress:
        text = grapheme.remove_stress(text)
    return text


def inflect_ordinal(
    whole: str,
    form: grammar.InflectionForm,
    stress: bool = False,
    insert_ones: bool = False,
    strict_range: bool = False,
) -> str:
    """
    Returns a spelling of the provided non-negative integer number as an ordinal
    numeral. The spelling is inflected for a requested form. The input is validated and
    exception of type descending from :any:`errors.YarNumBaseError` is thrown if
    validation fails. The range of numbers that can be correctly inflected is [0...1e27);
    numbers outside this range are not processed correctly. See `strict_range` for more
    details.

    :param whole: The string of digits in base 10 to be spelled out.
    :param form: The grammatical form of the generated text.
    :param stress: Whether the output text should contain stresses.
    :param insert_ones: Whether the output should insert "ones" in thousands and millions.
    :param strict_range: Whether to throw an exception when the number is too big to
    correctly process.
    :return: The spellout of the provided number as a ordinal numeral inflected for a given form.

    Basic usage
        >>> inflect_ordinal("43", {"case": "nominative"})
        'сорок третій'
        >>> inflect_ordinal("43", {"case": "genitive", "gender": "feminine"})
        'сорок третьої'

    Using `stress` parameter
        >>> inflect_ordinal("75", {}, stress=False)
        'сімдесят пʼятий'
        >>> inflect_ordinal("75", {}, stress=True)
        'сімдеся́т пʼя́тий'

    Using `insert_ones` parameter
        >>> inflect_ordinal("1001", {}, insert_ones=False)
        'тисяча перший'
        >>> inflect_ordinal("1001", {}, insert_ones=True)
        'одна тисяча перший'

    Using `strict_range` parameter
        >>> inflect_ordinal("100000000000000000000000000", {}, strict_range=False)
        'стасептильйонний'
        >>> inflect_ordinal("1000000000000000000000000000", {}, strict_range=False)
        'один нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нульовий'
        >>> inflect_ordinal("1000000000000000000000000000", {}, strict_range=True)
        Traceback (most recent call last):
            ...
        yar_numerals.errors.RangeError: Provided number is too large with 28 out of 27 digits.
    """
    validator.validate_int(whole)
    validator.validate_form(form)
    text = ordinals.inflect(whole, form, insert_ones, strict_range)
    if not stress:
        text = grapheme.remove_stress(text)
    return text


def inflect_decimal(
    whole: str,
    fractional: str,
    form: grammar.InflectionForm,
    stress: bool = False,
    insert_ones: bool = False,
    strict_range: bool = False,
) -> str:
    """
    Returns a spelling of the provided decimal fractional number as a cardinal
    numeral. The spelling is inflected for a requested form. The input is validated and
    exception of type descending from :any:`errors.YarNumBaseError` is thrown if
    validation fails. The number that can be correctly spelled only if every component
    of the fraction is in range [0...1e27); numbers with components outside of this
    range are not processed correctly. See `strict_range` for more details.

    :param whole: The string of digits in base 10 to be spelled out.
    :param fractional: The string of digits in base 10 to be spelled out.
    :param form: The grammatical form of the generated text.
    :param stress: Whether the output text should contain stresses.
    :param insert_ones: Whether the output should insert "ones" in thousands and millions.
    :param strict_range: Whether to throw an exception when the number is too big to
    correctly process.
    :return: The spellout of the provided number as a cardinal numeral inflected for a given form.

    Basic usage
        >>> inflect_cardinal("43", {"case": "nominative"})
        'сорок три'
        >>> inflect_cardinal("43", {"case": "genitive"})
        'сорока трьох'

    Using `stress` parameter
        >>> inflect_cardinal("75", {}, stress=False)
        'сімдесят пʼять'
        >>> inflect_cardinal("75", {}, stress=True)
        'сімдеся́т пʼя́ть'

    Using `insert_ones` parameter
        >>> inflect_cardinal("1000", {}, insert_ones=False)
        'тисяча'
        >>> inflect_cardinal("1000", {}, insert_ones=True)
        'одна тисяча'

    Using `strict_range` parameter
        >>> inflect_cardinal("100000000000000000000000000", {}, strict_range=False)
        'сто септильйонів'
        >>> inflect_cardinal("1000000000000000000000000000", {}, strict_range=False)
        'один нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль нуль'
        >>> inflect_cardinal("1000000000000000000000000000", {}, strict_range=True)
        Traceback (most recent call last):
            ...
        yar_numerals.errors.RangeError: Provided number is too large with 28 out of 27 digits.
    """
    validator.validate_int(whole)
    validator.validate_int(fractional)
    validator.validate_form(form)
    text = decimals.inflect(whole, fractional, form, insert_ones, strict_range)
    if not stress:
        text = grapheme.remove_stress(text)
    return text


def inflect_fractional(
    whole: str,
    numerator: str,
    denominator: str,
    form: grammar.InflectionForm,
    stress=False,
    insert_ones=False,
    strict_range=False,
) -> str:
    """
    Returns a spelling of the provided fractional number as a cardinal numeral.
    The spelling is inflected for a requested form. The input is validated and
    exception of type descending from :any:`errors.YarNumBaseError` is thrown if
    validation fails. The number that can be correctly spelled only if every component
    of the fraction is in range [0...1e27); numbers with components outside of this
    range are not processed correctly. See `strict_range` for more details.
    """
    validator.validate_int(whole)
    validator.validate_int(numerator)
    validator.validate_int(denominator)
    validator.validate_form(form)
    text = fractionals.inflect(
        whole, numerator, denominator, form, insert_ones, strict_range
    )
    if not stress:
        text = grapheme.remove_stress(text)
    return text



import re

from .errors import *
from .grammar import *

_intRegex = re.compile(r"\s*\d+\s*")


def validate_int(v: str):
    if not _intRegex.fullmatch(v):
        raise NonIntegerInputError(v)


def validate_form(f: InflectionForm):
    case = f.get("case")
    if case:
        if case not in list(Case):
            raise InvalidFormError(f"case has invalid value: {str(case)}")
    gender = f.get("gender")
    if gender:
        if gender not in list(Gender):
            raise InvalidFormError(f"gender has invalid value: {str(gender)}")
    number = f.get("number")
    if number:
        if number not in list(Number):
            raise InvalidFormError(f"number has invalid value: {str(number)}")
    animacy = f.get("animacy")
    if animacy:
        if animacy not in list(Animacy):
            raise InvalidFormError(f"animacy has invalid value: {str(animacy)}")



import pytest
from .cardinals import inflect
from .errors import RangeError


# fmt: off
simple_map = {
    "nom": ("case", "nominative"),
    "gen": ("case", "genitive"),
    "dat": ("case", "dative"),
    "acc": ("case", "accusative"),
    "inst": ("case", "instrumental"),
    "loc": ("case", "locative"),
    "voc": ("case", "vocative"),
    "m": ("gender", "masculine"),
    "f": ("gender", "feminine"),
    "n": ("gender", "neuter"),
    "sg": ("number", "singular"),
    "pl": ("number", "plural"),
    "card": ("type", "cardinal"),
    "ord": ("type", "ordinal"),
    "inan": ("animacy", "inanimate"),
    "anim": ("animacy", "animate"),
}


def fm(s):
    return {simple_map[g][0]: simple_map[g][1] for g in s.split(";")}


def test_0():
    assert inflect("0", fm("m")) == "ну́ль"
    assert inflect("0", fm("f")) == "ну́ль"
    assert inflect("0", fm("n")) == "ну́ль"
    assert inflect("0", fm("gen")) == "нуля́"
    assert inflect("0", fm("gen;f")) == "нуля́"
    assert inflect("0", fm("acc;f;anim")) == "ну́ль"
    assert inflect("0", fm("acc;f;inan")) == "ну́ль"
    assert inflect("0", fm("loc;f")) == "нулі́"


def test_lpad_0():
    assert inflect("00", fm("m")) == "ну́ль"
    assert inflect("000000", fm("m")) == "ну́ль"
    assert inflect("021", fm("m")) == "два́дцять оди́н"
    assert inflect("0000000021", fm("m")) == "два́дцять оди́н"


def test_21():
    assert inflect("21", fm("m")) == "два́дцять оди́н"
    assert inflect("21", fm("f")) == "два́дцять одна́"
    assert inflect("21", fm("n")) == "два́дцять одне́"
    assert inflect("21", fm("gen")) == "двадцяти́ одного́"
    assert inflect("21", fm("gen;f")) == "двадцяти́ одніє́ї"
    assert inflect("21", fm("acc;f;anim")) == "двадцятьо́х одну́"
    assert inflect("21", fm("acc;f;inan")) == "два́дцять одну́"
    assert inflect("21", fm("loc;f")) == "двадцяти́ одні́й"


def test_34():
    assert inflect("34", fm("m")) == "три́дцять чоти́ри"
    assert inflect("34", fm("f")) == "три́дцять чоти́ри"
    assert inflect("34", fm("n")) == "три́дцять чоти́ри"
    assert inflect("34", fm("gen")) == "тридцяти́ чотирьо́х"
    assert inflect("34", fm("gen;f")) == "тридцяти́ чотирьо́х"
    assert inflect("34", fm("acc;f;anim")) == "тридцятьо́х чотирьо́х"
    assert inflect("34", fm("acc;f;inan")) == "три́дцять чоти́ри"
    assert inflect("34", fm("loc;f")) == "тридцяти́ чотирьо́х"


def test_713():
    assert inflect("713", fm("m")) == "сімсо́т трина́дцять"
    assert inflect("713", fm("f")) == "сімсо́т трина́дцять"
    assert inflect("713", fm("n")) == "сімсо́т трина́дцять"
    assert inflect("713", fm("gen")) == "семисо́т тринадцяти́"
    assert inflect("713", fm("gen;f")) == "семисо́т тринадцяти́"
    assert inflect("713", fm("acc;f;anim")) == "сімсо́т тринадцятьо́х"
    assert inflect("713", fm("acc;f;inan")) == "сімсо́т трина́дцять"
    assert inflect("713", fm("loc;f")) == "семиста́х тринадцяти́"


def test_999():
    assert inflect("999", fm("m")) == "девʼятсо́т девʼяно́сто де́вʼять"
    assert inflect("999", fm("f")) == "девʼятсо́т девʼяно́сто де́вʼять"
    assert inflect("999", fm("n")) == "девʼятсо́т девʼяно́сто де́вʼять"
    assert inflect("999", fm("gen")) == "девʼятисо́т девʼяно́ста девʼяти́"
    assert inflect("999", fm("gen;f")) == "девʼятисо́т девʼяно́ста девʼяти́"
    assert inflect("999", fm("acc;f;anim")) == "девʼятсо́т девʼяно́сто девʼятьо́х"
    assert inflect("999", fm("acc;f;inan")) == "девʼятсо́т девʼяно́сто де́вʼять"
    assert inflect("999", fm("loc;f")) == "девʼятиста́х девʼяно́ста девʼяти́"


def test_1000():
    assert inflect("1000", fm("m")) == "ти́сяча"
    assert inflect("1000", fm("f")) == "ти́сяча"
    assert inflect("1000", fm("n")) == "ти́сяча"
    assert inflect("1000", fm("gen;m")) == "ти́сячі"
    assert inflect("1000", fm("gen;f")) == "ти́сячі"
    assert inflect("1000", fm("dat")) == "ти́сячі"
    assert inflect("1000", fm("acc")) == "ти́сячу"
    assert inflect("1000", fm("acc;f;anim")) == "ти́сячу"
    assert inflect("1000", fm("acc;f;inan")) == "ти́сячу"
    assert inflect("1000", fm("loc;f")) == "ти́сячі"
    assert inflect("1000", fm("loc")) == "ти́сячі"
    assert inflect("1000", fm("inst")) == "ти́сячею"


def test_1000_insert_ones():
    assert inflect("1000", fm("m"), insert_ones=True) == "одна́ ти́сяча"
    assert inflect("1000", fm("f"), insert_ones=True) == "одна́ ти́сяча"
    assert inflect("1000", fm("n"), insert_ones=True) == "одна́ ти́сяча"
    assert inflect("1000", fm("gen;m"), insert_ones=True) == "одніє́ї ти́сячі"
    assert inflect("1000", fm("gen;f"), insert_ones=True) == "одніє́ї ти́сячі"
    assert inflect("1000", fm("dat"), insert_ones=True) == "одні́й ти́сячі"
    assert inflect("1000", fm("acc"), insert_ones=True) == "одну́ ти́сячу"
    assert inflect("1000", fm("acc;f;anim"), insert_ones=True) == "одну́ ти́сячу"
    assert inflect("1000", fm("acc;f;inan"), insert_ones=True) == "одну́ ти́сячу"
    assert inflect("1000", fm("loc;f"), insert_ones=True) == "одні́й ти́сячі"
    assert inflect("1000", fm("loc"), insert_ones=True) == "одні́й ти́сячі"
    assert inflect("1000", fm("inst"), insert_ones=True) == "одніє́ю ти́сячею"


def test_4486():
    assert inflect("4486", fm("m")) == "чоти́ри ти́сячі чоти́риста вісімдеся́т ші́сть"
    assert inflect("4486", fm("f")) == "чоти́ри ти́сячі чоти́риста вісімдеся́т ші́сть"
    assert inflect("4486", fm("n")) == "чоти́ри ти́сячі чоти́риста вісімдеся́т ші́сть"
    assert inflect("4486", fm("gen")) == "чотирьо́х ти́сяч чотирьохсо́т вісімдесяти́ шести́"
    assert inflect("4486", fm("gen;f")) == "чотирьо́х ти́сяч чотирьохсо́т вісімдесяти́ шести́"
    assert inflect("4486", fm("acc;f;anim")) == "чоти́ри ти́сячі чоти́риста вісімдесятьо́х шістьо́х"
    assert inflect("4486", fm("acc;f;inan")) == "чоти́ри ти́сячі чоти́риста вісімдеся́т ші́сть"
    assert inflect("4486", fm("loc;f")) == "чотирьо́х ти́сячах чотирьохста́х вісімдесяти́ шести́"


def test_1000000():
    assert inflect("1000000", fm("nom")) == "мільйо́н"
    assert inflect("1000000", fm("gen")) == "мільйо́на"
    assert inflect("1000000", fm("dat")) == "мільйо́ну"
    assert inflect("1000000", fm("acc")) == "мільйо́н"
    assert inflect("1000000", fm("inst")) == "мільйо́ном"
    assert inflect("1000000", fm("loc")) == "мільйо́ні"


def test_1000000_insert_ones():
    assert inflect("1000000", fm("nom"), insert_ones=True) == "оди́н мільйо́н"
    assert inflect("1000000", fm("gen"), insert_ones=True) == "одного́ мільйо́на"
    assert inflect("1000000", fm("dat"), insert_ones=True) == "одному́ мільйо́ну"
    assert inflect("1000000", fm("acc"), insert_ones=True) == "оди́н мільйо́н"
    assert inflect("1000000", fm("inst"), insert_ones=True) == "одни́м мільйо́ном"
    assert inflect("1000000", fm("loc"), insert_ones=True) == "одному́ мільйо́ні"


def test_801066():
    assert inflect("801066", fm("nom")) == "вісімсо́т одна́ ти́сяча шістдеся́т ші́сть"
    assert inflect("801066", fm("gen")) == "восьмисо́т одніє́ї ти́сячі шістдесяти́ шести́"
    assert inflect("801066", fm("dat")) == "восьмиста́м одні́й ти́сячі шістдесяти́ шести́"
    assert inflect("801066", fm("acc")) == "вісімсо́т одну́ ти́сячу шістдеся́т ші́сть"
    assert inflect("801066", fm("inst")) == "вісьмаста́ми одніє́ю ти́сячею шістдесятьма́ шістьма́"
    assert inflect("801066", fm("loc")) == "восьмиста́х одні́й ти́сячі шістдесяти́ шести́"


def test_5178030():
    assert inflect("5178030", fm("m")) == "пʼя́ть мільйо́нів сто́ сімдеся́т ві́сім ти́сяч три́дцять"
    assert inflect("5178030", fm("f")) == "пʼя́ть мільйо́нів сто́ сімдеся́т ві́сім ти́сяч три́дцять"
    assert inflect("5178030", fm("n")) == "пʼя́ть мільйо́нів сто́ сімдеся́т ві́сім ти́сяч три́дцять"
    assert inflect("5178030", fm("gen")) == "пʼяти́ мільйо́нів ста́ сімдесяти́ восьми́ ти́сяч тридцяти́"
    assert inflect("5178030", fm("gen;f")) == "пʼяти́ мільйо́нів ста́ сімдесяти́ восьми́ ти́сяч тридцяти́"
    assert inflect("5178030", fm("acc;f;anim")) == "пʼя́ть мільйо́нів сто́ сімдеся́т ві́сім ти́сяч тридцятьо́х"
    assert inflect("5178030", fm("acc;f;inan")) == "пʼя́ть мільйо́нів сто́ сімдеся́т ві́сім ти́сяч три́дцять"
    assert inflect("5178030", fm("loc;f")) == "пʼяти́ мільйо́нах ста́ сімдесяти́ восьми́ ти́сячах тридцяти́"


def test_488108826177873():
    assert inflect("488108826177873", fm("m")) == "чоти́риста вісімдеся́т ві́сім трильйо́нів сто́ ві́сім мілья́рдів вісімсо́т два́дцять ші́сть мільйо́нів сто́ сімдеся́т сі́м ти́сяч вісімсо́т сімдеся́т три́"
    assert inflect("488108826177873", fm("f")) == "чоти́риста вісімдеся́т ві́сім трильйо́нів сто́ ві́сім мілья́рдів вісімсо́т два́дцять ші́сть мільйо́нів сто́ сімдеся́т сі́м ти́сяч вісімсо́т сімдеся́т три́"
    assert inflect("488108826177873", fm("n")) == "чоти́риста вісімдеся́т ві́сім трильйо́нів сто́ ві́сім мілья́рдів вісімсо́т два́дцять ші́сть мільйо́нів сто́ сімдеся́т сі́м ти́сяч вісімсо́т сімдеся́т три́"
    assert inflect("488108826177873", fm("gen")) == "чотирьохсо́т вісімдесяти́ восьми́ трильйо́нів ста́ восьми́ мілья́рдів восьмисо́т двадцяти́ шести́ мільйо́нів ста́ сімдесяти́ семи́ ти́сяч восьмисо́т сімдесяти́ трьо́х"
    assert inflect("488108826177873", fm("gen;f")) == "чотирьохсо́т вісімдесяти́ восьми́ трильйо́нів ста́ восьми́ мілья́рдів восьмисо́т двадцяти́ шести́ мільйо́нів ста́ сімдесяти́ семи́ ти́сяч восьмисо́т сімдесяти́ трьо́х", "genitive, feminine"
    assert inflect("488108826177873", fm("acc;f;anim")) == "чоти́риста вісімдеся́т ві́сім трильйо́нів сто́ ві́сім мілья́рдів вісімсо́т два́дцять ші́сть мільйо́нів сто́ сімдеся́т сі́м ти́сяч вісімсо́т сімдесятьо́х трьо́х"
    assert inflect("488108826177873", fm("acc;f;inan")) == "чоти́риста вісімдеся́т ві́сім трильйо́нів сто́ ві́сім мілья́рдів вісімсо́т два́дцять ші́сть мільйо́нів сто́ сімдеся́т сі́м ти́сяч вісімсо́т сімдеся́т три́"
    assert inflect("488108826177873", fm("loc;f")) == "чотирьохста́х вісімдесяти́ восьми́ трильйо́нах ста́ восьми́ мілья́рдах восьмиста́х двадцяти́ шести́ мільйо́нах ста́ сімдесяти́ семи́ ти́сячах восьмиста́х сімдесяти́ трьо́х", "locative, feminine"


def test_591359027947837343695787388():
    assert inflect("591359027947837343695787388", fm("nom;m")) == "пʼятсо́т девʼяно́сто оди́н септильйо́н три́ста пʼятдеся́т де́вʼять секстильйо́нів два́дцять сі́м квінтильйо́нів девʼятсо́т со́рок сі́м квадрильйо́нів вісімсо́т три́дцять сі́м трильйо́нів три́ста со́рок три́ мілья́рди шістсо́т девʼяно́сто пʼя́ть мільйо́нів сімсо́т вісімдеся́т сі́м ти́сяч три́ста вісімдеся́т ві́сім"


def test_nominative():
    assert inflect("0", fm("nom")) == "ну́ль"
    assert inflect("1", fm("nom")) == "оди́н"
    assert inflect("4", fm("nom")) == "чоти́ри"
    assert inflect("5", fm("nom")) == "пʼя́ть"
    assert inflect("6", fm("nom")) == "ші́сть"
    assert inflect("8", fm("nom")) == "ві́сім"
    assert inflect("9", fm("nom")) == "де́вʼять"
    assert inflect("10", fm("nom")) == "де́сять"
    assert inflect("11", fm("nom")) == "одина́дцять"
    assert inflect("15", fm("nom")) == "пʼятна́дцять"
    assert inflect("16", fm("nom")) == "шістна́дцять"
    assert inflect("18", fm("nom")) == "вісімна́дцять"
    assert inflect("19", fm("nom")) == "девʼятна́дцять"
    assert inflect("20", fm("nom")) == "два́дцять"
    assert inflect("30", fm("nom")) == "три́дцять"
    assert inflect("100", fm("nom")) == "сто́"


def test_genitive():
    assert inflect("0", fm("gen")) == "нуля́"
    assert inflect("1", fm("gen")) == "одного́"
    assert inflect("2", fm("gen")) == "дво́х"
    assert inflect("3", fm("gen")) == "трьо́х"
    assert inflect("4", fm("gen")) == "чотирьо́х"
    assert inflect("5", fm("gen")) == "пʼяти́"
    assert inflect("6", fm("gen")) == "шести́"
    assert inflect("7", fm("gen")) == "семи́"
    assert inflect("8", fm("gen")) == "восьми́"
    assert inflect("9", fm("gen")) == "девʼяти́"
    assert inflect("10", fm("gen")) == "десяти́"
    assert inflect("11", fm("gen")) == "одинадцяти́"
    assert inflect("12", fm("gen")) == "дванадцяти́"
    assert inflect("13", fm("gen")) == "тринадцяти́"
    assert inflect("14", fm("gen")) == "чотирнадцяти́"
    assert inflect("15", fm("gen")) == "пʼятнадцяти́"
    assert inflect("16", fm("gen")) == "шістнадцяти́"
    assert inflect("17", fm("gen")) == "сімнадцяти́"
    assert inflect("18", fm("gen")) == "вісімнадцяти́"
    assert inflect("19", fm("gen")) == "девʼятнадцяти́"
    assert inflect("20", fm("gen")) == "двадцяти́"
    assert inflect("30", fm("gen")) == "тридцяти́"
    assert inflect("100", fm("gen")) == "ста́"


def test_dative():
    assert inflect("0", fm("dat")) == "нулю́"
    assert inflect("1", fm("dat")) == "одному́"
    assert inflect("4", fm("dat")) == "чотирьо́м"
    assert inflect("5", fm("dat")) == "пʼяти́"
    assert inflect("6", fm("dat")) == "шести́"
    assert inflect("8", fm("dat")) == "восьми́"
    assert inflect("9", fm("dat")) == "девʼяти́"
    assert inflect("10", fm("dat")) == "десяти́"
    assert inflect("11", fm("dat")) == "одинадцяти́"
    assert inflect("15", fm("dat")) == "пʼятнадцяти́"
    assert inflect("16", fm("dat")) == "шістнадцяти́"
    assert inflect("18", fm("dat")) == "вісімнадцяти́"
    assert inflect("19", fm("dat")) == "девʼятнадцяти́"
    assert inflect("20", fm("dat")) == "двадцяти́"
    assert inflect("30", fm("dat")) == "тридцяти́"
    assert inflect("100", fm("dat")) == "ста́"


def test_accusative():
    assert inflect("0", fm("acc")) == "ну́ль"
    assert inflect("1", fm("acc")) == "оди́н"
    assert inflect("4", fm("acc")) == "чоти́ри"
    assert inflect("5", fm("acc")) == "пʼя́ть"
    assert inflect("6", fm("acc")) == "ші́сть"
    assert inflect("8", fm("acc")) == "ві́сім"
    assert inflect("9", fm("acc")) == "де́вʼять"
    assert inflect("10", fm("acc")) == "де́сять"
    assert inflect("11", fm("acc")) == "одина́дцять"
    assert inflect("15", fm("acc")) == "пʼятна́дцять"
    assert inflect("16", fm("acc")) == "шістна́дцять"
    assert inflect("18", fm("acc")) == "вісімна́дцять"
    assert inflect("19", fm("acc")) == "девʼятна́дцять"
    assert inflect("20", fm("acc")) == "два́дцять"
    assert inflect("30", fm("acc")) == "три́дцять"
    assert inflect("100", fm("acc")) == "сто́"


def test_instrumental():
    assert inflect("0", fm("inst")) == "нуле́м"
    assert inflect("1", fm("inst")) == "одни́м"
    assert inflect("4", fm("inst")) == "чотирма́"
    assert inflect("5", fm("inst")) == "пʼятьма́"
    assert inflect("6", fm("inst")) == "шістьма́"
    assert inflect("8", fm("inst")) == "вісьма́"
    assert inflect("9", fm("inst")) == "девʼятьма́"
    assert inflect("10", fm("inst")) == "десятьма́"
    assert inflect("11", fm("inst")) == "одинадцятьма́"
    assert inflect("15", fm("inst")) == "пʼятнадцятьма́"
    assert inflect("16", fm("inst")) == "шістнадцятьма́"
    assert inflect("18", fm("inst")) == "вісімнадцятьма́"
    assert inflect("19", fm("inst")) == "девʼятнадцятьма́"
    assert inflect("20", fm("inst")) == "двадцятьма́"
    assert inflect("30", fm("inst")) == "тридцятьма́"
    assert inflect("100", fm("inst")) == "ста́"


def test_locative():
    assert inflect("0", fm("loc")) == "нулі́"
    assert inflect("1", fm("loc")) == "одному́"
    assert inflect("4", fm("loc")) == "чотирьо́х"
    assert inflect("5", fm("loc")) == "пʼяти́"
    assert inflect("6", fm("loc")) == "шести́"
    assert inflect("8", fm("loc")) == "восьми́"
    assert inflect("9", fm("loc")) == "девʼяти́"
    assert inflect("10", fm("loc")) == "десяти́"
    assert inflect("11", fm("loc")) == "одинадцяти́"
    assert inflect("15", fm("loc")) == "пʼятнадцяти́"
    assert inflect("16", fm("loc")) == "шістнадцяти́"
    assert inflect("19", fm("loc")) == "девʼятнадцяти́"
    assert inflect("20", fm("loc")) == "двадцяти́"
    assert inflect("30", fm("loc")) == "тридцяти́"
    assert inflect("100", fm("loc")) == "ста́"



def test_vocative():
    assert inflect("0", fm("voc")) == "ну́лю"
    assert inflect("1", fm("voc")) == "оди́н"
    assert inflect("4", fm("voc")) == "чоти́ри"
    assert inflect("5", fm("voc")) == "пʼя́ть"
    assert inflect("6", fm("voc")) == "ші́сть"
    assert inflect("8", fm("voc")) == "ві́сім"
    assert inflect("9", fm("voc")) == "де́вʼять"
    assert inflect("10", fm("voc")) == "де́сять"
    assert inflect("11", fm("voc")) == "одина́дцять"
    assert inflect("15", fm("voc")) == "пʼятна́дцять"
    assert inflect("16", fm("voc")) == "шістна́дцять"
    assert inflect("18", fm("voc")) == "вісімна́дцять"
    assert inflect("19", fm("voc")) == "девʼятна́дцять"
    assert inflect("20", fm("voc")) == "два́дцять"
    assert inflect("30", fm("voc")) == "три́дцять"
    assert inflect("100", fm("voc")) == "сто́"

def test_all_neuter():
    f = fm("inst;n")
    assert inflect("1", f) == "одни́м"
    assert inflect("2", f) == "двома́"
    assert inflect("3", f) == "трьома́"
    assert inflect("4", f) == "чотирма́"
    assert inflect("5", f) == "пʼятьма́"
    assert inflect("6", f) == "шістьма́"
    assert inflect("7", f) == "сьома́"
    assert inflect("8", f) == "вісьма́"
    assert inflect("9", f) == "девʼятьма́"
    assert inflect("11", f) == "одинадцятьма́"
    assert inflect("12", f) == "дванадцятьма́"
    assert inflect("13", f) == "тринадцятьма́"
    assert inflect("14", f) == "чотирнадцятьма́"
    assert inflect("15", f) == "пʼятнадцятьма́"
    assert inflect("16", f) == "шістнадцятьма́"
    assert inflect("17", f) == "сімнадцятьма́"
    assert inflect("18", f) == "вісімнадцятьма́"
    assert inflect("19", f) == "девʼятнадцятьма́"
    assert inflect("10", f) == "десятьма́"
    assert inflect("20", f) == "двадцятьма́"
    assert inflect("30", f) == "тридцятьма́"
    assert inflect("40", f) == "сорока́"
    assert inflect("50", f) == "пʼятдесятьма́"
    assert inflect("60", f) == "шістдесятьма́"
    assert inflect("70", f) == "сімдесятьма́"
    assert inflect("80", f) == "вісімдесятьма́"
    assert inflect("90", f) == "девʼяно́ста"
    assert inflect("100", f) == "ста́"
    assert inflect("200", f) == "двомаста́ми"
    assert inflect("300", f) == "трьомаста́ми"
    assert inflect("400", f) == "чотирмаста́ми"
    assert inflect("500", f) == "пʼятьмаста́ми"
    assert inflect("600", f) == "шістьмаста́ми"
    assert inflect("700", f) == "семиста́ми"
    assert inflect("800", f) == "вісьмаста́ми"
    assert inflect("900", f) == "девʼятьмаста́ми"


def test_range_fallback():
    assert inflect("999999999999999999999999999", fm("m")) == "девʼятсо́т девʼяно́сто де́вʼять септильйо́нів девʼятсо́т девʼяно́сто де́вʼять секстильйо́нів девʼятсо́т девʼяно́сто де́вʼять квінтильйо́нів девʼятсо́т девʼяно́сто де́вʼять квадрильйо́нів девʼятсо́т девʼяно́сто де́вʼять трильйо́нів девʼятсо́т девʼяно́сто де́вʼять мілья́рдів девʼятсо́т девʼяно́сто де́вʼять мільйо́нів девʼятсо́т девʼяно́сто де́вʼять ти́сяч девʼятсо́т девʼяно́сто де́вʼять"
    assert inflect("1000000000000000000000000000", fm("m")) == "оди́н ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль"
    assert inflect("1000000000000000000000000001", fm("m")) == "оди́н ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль оди́н"
    assert inflect("1234567890123456789012345678", fm("m")) == "оди́н два́ три́ чоти́ри пʼя́ть ші́сть сі́м ві́сім де́вʼять ну́ль оди́н два́ три́ чоти́ри пʼя́ть ші́сть сі́м ві́сім де́вʼять ну́ль оди́н два́ три́ чоти́ри пʼя́ть ші́сть сі́м ві́сім"
    assert inflect("999999999999999999999999999", fm("f")) == "девʼятсо́т девʼяно́сто де́вʼять септильйо́нів девʼятсо́т девʼяно́сто де́вʼять секстильйо́нів девʼятсо́т девʼяно́сто де́вʼять квінтильйо́нів девʼятсо́т девʼяно́сто де́вʼять квадрильйо́нів девʼятсо́т девʼяно́сто де́вʼять трильйо́нів девʼятсо́т девʼяно́сто де́вʼять мілья́рдів девʼятсо́т девʼяно́сто де́вʼять мільйо́нів девʼятсо́т девʼяно́сто де́вʼять ти́сяч девʼятсо́т девʼяно́сто де́вʼять"
    assert inflect("1000000000000000000000000000", fm("f")) == "оди́н ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль"
    assert inflect("1000000000000000000000000001", fm("f")) == "оди́н ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль одна́"
    assert inflect("1234567890123456789012345678", fm("f")) == "оди́н два́ три́ чоти́ри пʼя́ть ші́сть сі́м ві́сім де́вʼять ну́ль оди́н два́ три́ чоти́ри пʼя́ть ші́сть сі́м ві́сім де́вʼять ну́ль оди́н два́ три́ чоти́ри пʼя́ть ші́сть сі́м ві́сім"
    assert inflect("1000000000000000000000000000", fm("gen;f")) == "оди́н ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль нуля́"
    assert inflect("1000000000000000000000000001", fm("gen;f")) == "оди́н ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль одніє́ї"


def test_strict_range():
    assert inflect("999999999999999999999999999", fm("m"), strict_range=True)
    with pytest.raises(RangeError):
        assert inflect("1000000000000000000000000000", fm("m"), strict_range=True)


def test_insert_ones():
    assert inflect("100", fm("m"), insert_ones=True) == "сто́"
    assert inflect("101", fm("m"), insert_ones=True) == "сто́ оди́н"
    assert inflect("1000", fm("m"), insert_ones=True) == "одна́ ти́сяча"
    assert inflect("1001", fm("m"), insert_ones=True) == "одна́ ти́сяча оди́н"
    assert inflect("1000000", fm("m"), insert_ones=True) == "оди́н мільйо́н"
    assert inflect("1000001", fm("m"), insert_ones=True) == "оди́н мільйо́н оди́н"
    assert inflect("100", fm("m"), insert_ones=True) == "сто́"
    assert inflect("101", fm("f"), insert_ones=True) == "сто́ одна́"
    assert inflect("1000", fm("f"), insert_ones=True) == "одна́ ти́сяча"
    assert inflect("1001", fm("f"), insert_ones=True) == "одна́ ти́сяча одна́"
    assert inflect("1000000", fm("f"), insert_ones=True) == "оди́н мільйо́н"
    assert inflect("1000001", fm("f"), insert_ones=True) == "оди́н мільйо́н одна́"


"""
Numbers to also test:
2
3
7
12
13
14
42
71
28
31
35
51
38
71
101
104
105
109
1001
4486
9999
100000
803066
999999
1000001
9999999
100000000
857049589
999999999
1000000000
1000000000000
488108826177873
999999999999999
"""

# fmt: on



# (c) Copyright by Yar.
#
# Source file generated automatically.
# fmt: off


data = {
  "whole": ([
    (dict(gender="masculine", number="singular", case="nominative"), "ці́лий"),
    (dict(gender="feminine", number="singular", case="nominative"), "ці́ла"),
    (dict(gender="neuter", number="singular", case="nominative"), "ці́ле"),
    (dict(number="plural", case="nominative"), "ці́лі"),
    (dict(gender="masculine", number="singular", case="genitive"), "ці́лого"),
    (dict(gender="feminine", number="singular", case="genitive"), "ці́лої"),
    (dict(gender="neuter", number="singular", case="genitive"), "ці́лого"),
    (dict(number="plural", case="genitive"), "ці́лих"),
    (dict(gender="masculine", number="singular", case="dative"), "ці́лому"),
    (dict(gender="feminine", number="singular", case="dative"), "ці́лій"),
    (dict(gender="neuter", number="singular", case="dative"), "ці́лому"),
    (dict(number="plural", case="dative"), "ці́лим"),
    (dict(gender="masculine", number="singular", case="accusative", animacy="inanimate"), "ці́лий"),
    (dict(gender="masculine", number="singular", case="accusative", animacy="animate"), "ці́лого"),
    (dict(gender="feminine", number="singular", case="accusative"), "ці́лу"),
    (dict(gender="neuter", number="singular", case="accusative"), "ці́ле"),
    (dict(number="plural", case="accusative", animacy="inanimate"), "ці́лі"),
    (dict(number="plural", case="accusative", animacy="animate"), "ці́лих"),
    (dict(gender="masculine", number="singular", case="instrumental"), "ці́лим"),
    (dict(gender="feminine", number="singular"), "ці́лою"),
    (dict(gender="neuter", number="singular", case="instrumental"), "ці́лим"),
    (dict(number="plural", case="instrumental"), "ці́лими"),
    (dict(gender="masculine", number="singular", case="locative"), "ці́лому"),
    (dict(gender="masculine", number="singular", case="locative"), "ці́лім"),
    (dict(gender="feminine", number="singular"), "ці́лій"),
    (dict(gender="neuter", number="singular", case="locative"), "ці́лому"),
    (dict(gender="neuter", number="singular", case="locative"), "ці́лім"),
    (dict(number="plural", case="locative"), "ці́лих"),
  ], dict()),
}



from .grammar import *
from .syntax import *
from . import cardinals, ordinals
from .lexeme import NumeralLexeme


def inflect(
    whole: str,
    numerator: str,
    denominator: str,
    form: InflectionForm,
    insert_ones=False,
    strict_range=False,
) -> str:
    root = DummySyntaxNode(form)
    obj = DummySyntaxNode({})

    build_tree(root, obj, whole, numerator, denominator, insert_ones, strict_range)
    root.agree()
    return str(root)


def build_tree(
    root: SyntaxNode,
    obj: SyntaxNode,
    whole: str,
    numerator: str,
    denominator: str,
    insert_ones=False,
    strict_range=False,
) -> None:
    # Add a dummy node for the virtual lexeme "частина" as in
    #  1/4 = одна четверта [частина]
    vroot = DummySyntaxNode(
        persistent_form=InflectionForm(gender="feminine", animacy="inanimate"),
    )
    root.add_child_node(vroot, Relation.amod)
    vroot.add_child_node(obj, Relation.nmod)
    ordinals.build_tree(vroot, denominator, insert_ones, strict_range)
    cardinals.build_tree(vroot, numerator, insert_ones, strict_range)
    # TODO: find the right solutino
    _fix_puacal(vroot)
    if whole:
        whole_adj = vroot.add_child(NumeralLexeme.misc("whole"), Relation.amod)
        cardinals.build_tree(vroot, whole, insert_ones, strict_range)


def _fix_puacal(root: SyntaxNode):
    for edge in root.edges:
        if edge.rel == Relation.nummod_govpc:
            edge.rel = Relation.nummod_govpl



from .grammar import *
from .syntax import *
from . import fractionals


def inflect(
    whole: str,
    decimal: str,
    form: InflectionForm,
    insert_ones=False,
    strict_range=False,
) -> str:
    magnitude = get_decimal_fraction_magnitude(decimal)

    root = DummySyntaxNode(form)
    obj = DummySyntaxNode({})

    fractionals.build_tree(
        root, obj, whole, decimal, magnitude, insert_ones, strict_range
    )
    root.agree()
    return str(root)


def get_decimal_fraction_magnitude(decimal: str) -> str:
    return "1" + "0" * len(decimal)



from typing import List
from .grammar import *
from .lexeme import *
from .syntax import *
from . import cardinals, errors

MAX_DIGIT_COUNT = 27


def inflect(
    whole: str, form: InflectionForm, insert_ones=False, strict_range=False
) -> str:
    root = DummySyntaxNode(form)
    build_tree(root, whole, insert_ones, strict_range)
    root.agree()
    return str(root)


def build_tree(
    root: SyntaxNode, number: str, insert_ones: bool = False, strict_range: bool = False
) -> None:
    number = number.lstrip("0")
    if len(number) == 0:
        root.add_child(NumeralLexeme.ordinal("0"), Relation.nummod)
        return
    digits = list(number)
    build_tree_from_digits(root, digits, insert_ones, strict_range)


def build_tree_from_digits(
    root: SyntaxNode, digits: List[str], insert_ones: bool, strict_range: bool
):
    if len(digits) > MAX_DIGIT_COUNT:
        if strict_range:
            raise errors.RangeError(digits, MAX_DIGIT_COUNT)
        else:
            return build_tree_digit_wise(root, digits)

    power = 0
    while len(digits) > 0:
        [*digits, hundreds, tens, ones] = ["0", "0", *digits]
        numbers = list(triplet_into_monolexemic_numbers(hundreds, tens, ones))

        if len(numbers) > 0:
            if power == 0:
                [ord, *card] = numbers
                root.add_child(NumeralLexeme.ordinal(ord), Relation.amod)

                for n in card:
                    root.add_child(NumeralLexeme.cardinal(n), Relation.num)
                break
            else:
                large_number = "1" + "0" * power
                if numbers == ["1"]:
                    root.add_child(
                        NumeralLexeme.ordinal(large_number),
                        Relation.amod,
                    )
                else:
                    compound_lexeme = NumeralLexeme.ordinal_compound(
                        numbers, large_number
                    )
                    root.add_child(compound_lexeme, Relation.nummod)
                break

        power += 3
    power += 3

    # The remaining parts should be inflected as a cardinal
    is_first = False
    while len(digits) > 0:
        [*digits, hundreds, tens, ones] = ["0", "0", *digits]
        lexemes = list(cardinals.triplet_into_lexemes(hundreds, tens, ones))
        if len(lexemes) > 0:
            countable_lexeme = NumeralLexeme.cardinal("1" + "0" * power)
            current_root = root.add_child(countable_lexeme, Relation.num)
            is_first = True

            if not insert_ones:
                if len(lexemes) == 1 and lexemes[0].value == "1":
                    continue
            for lexeme in lexemes:
                current_root.add_child(
                    lexeme, cardinals.get_relation_type(lexeme, is_first)
                )
                is_first = False
        power += 3


def build_tree_digit_wise(root: SyntaxNode, digits: List[str]) -> None:
    [*cardinal_parts, ordinal_part] = digits
    root.add_child(NumeralLexeme.ordinal(ordinal_part), Relation.nummod)
    for digit in cardinal_parts[::-1]:
        root.add_child(NumeralLexeme.cardinal(digit), Relation.num)


def triplet_into_monolexemic_numbers(hundreds, tens, ones):
    if tens == "1":
        yield tens + ones
    else:
        if ones != "0":
            yield ones
        if tens != "0":
            yield tens + "0"
    if hundreds != "0":
        yield hundreds + "00"



from typing import List
from .grammar import *
from .lexeme import *
from .syntax import *
from . import errors

MAX_DIGIT_COUNT = 27


def inflect(
    whole: str,
    form: InflectionForm,
    insert_ones: bool = False,
    strict_range: bool = False,
) -> str:
    """
    Produce an inflected spelling of the provided whole cardinal number.

    *Note:* Input values are not validated, the recommended API is :any:`inflect_cardinal`

    :param whole: The sequence of base 10 digits to be spelled out.
    :param form: The grammatical form of the generated text.
    :param insert_ones: Whether the output should insert "ones" in thousands and millions.
    :param strict_range: Whether to throw an exception when the number is too big to correctly process.
    :return: The spellout of the provided whole cardinal number in a given form.
    """
    root = DummySyntaxNode(form)
    build_tree(root, whole, insert_ones, strict_range)
    root.agree()
    return str(root)


def build_tree(
    root: SyntaxNode, number: str, insert_ones: bool = False, strict_range: bool = False
) -> None:
    number = number.lstrip("0")
    if len(number) == 0:
        root.add_child(NumeralLexeme.cardinal("0"), Relation.nummod_govpl)
        return
    digits = list(number)
    build_tree_from_digits(root, digits, insert_ones, strict_range)


def build_tree_from_digits(
    root: SyntaxNode, digits: List[str], insert_ones: bool, strict_range: bool
) -> None:
    if len(digits) > MAX_DIGIT_COUNT:
        if strict_range:
            raise errors.RangeError(digits, MAX_DIGIT_COUNT)
        else:
            return build_tree_digit_wise(root, digits)

    is_first = True
    power = 0

    while len(digits) > 0:
        [*digits, hundreds, tens, ones] = ["0", "0", *digits]
        lexemes = list(triplet_into_lexemes(hundreds, tens, ones))
        if len(lexemes) > 0:
            current_root = root
            if power > 0:
                countable_lexeme = NumeralLexeme.cardinal("1" + "0" * power)
                current_root = root.add_child(
                    countable_lexeme, get_relation_type(countable_lexeme, is_first)
                )
                is_first = True

            if not insert_ones and power > 0:
                if len(lexemes) == 1 and lexemes[0].value == "1":
                    continue
            for lexeme in lexemes:
                current_root.add_child(lexeme, get_relation_type(lexeme, is_first))
                is_first = False
        power += 3


def build_tree_digit_wise(root, digits):
    [*remaining_parts, inflected_part] = digits
    root.add_child(NumeralLexeme.cardinal(inflected_part), Relation.nummod)
    for digit in remaining_parts[::-1]:
        root.add_child(NumeralLexeme.cardinal(digit), Relation.num)
    return


def get_relation_type(lexeme, is_first):
    if not is_first:
        return Relation.nummod
    if lexeme.value == "1":
        return Relation.nummod_govsg
    if lexeme.value in ("2", "3", "4"):
        return Relation.nummod_govpc
    else:
        return Relation.nummod_govpl


def triplet_into_lexemes(hundreds, tens, ones):
    if tens == "1":
        yield NumeralLexeme.cardinal(tens + ones)
    else:
        if ones != "0":
            yield NumeralLexeme.cardinal(ones)
        if tens != "0":
            yield NumeralLexeme.cardinal(tens + "0")
    if hundreds != "0":
        yield NumeralLexeme.cardinal(hundreds + "00")



# (c) Copyright by Yar.
#
# Syntactical layer of processing.

from typing import List, Iterable
from ..lexeme import *
from ..grammar import *
from .relation import *

__all__ = [
    "Relation",
    "SyntaxNode",
    "DummySyntaxNode",
]


class SyntaxNode:
    def __init__(self, lexeme: Lexeme) -> None:
        self.lexeme = lexeme
        self.edges: List["SyntaxEdge"] = []

    def add_child(self, lexeme: Lexeme, rel: Relation) -> "SyntaxNode":
        child = SyntaxNode(lexeme)
        return self.add_child_node(child, rel)

    def add_child_node(self, child: "SyntaxNode", rel: Relation) -> "SyntaxNode":
        edge = SyntaxEdge(self, child, rel)
        self.edges.append(edge)
        return child

    def agree(self) -> None:
        for edge in self.edges[::-1]:
            edge.agree()

    def lexemes(self) -> Iterable[Lexeme]:
        yield self.lexeme
        for edge in self.edges:
            yield from edge.child.lexemes()

    def __str__(self):
        return " ".join([l.text() for l in self.lexemes()][::-1])

    def __repr__(self):  # pragma: no cover
        if len(self.edges) == 0:
            return f"SyntaxNode<{repr(self.lexeme)}: []>"
        if len(self.edges) == 1:
            return f"SyntaxNode<{repr(self.lexeme)}: [{repr(self.edges[0])}]>"
        return f"SyntaxNode<{repr(self.lexeme)}: [{len(self.edges)}]>"


class DummySyntaxNode(SyntaxNode):
    def __init__(
        self, form: InflectionForm = None, persistent_form: InflectionForm = None
    ) -> None:
        if persistent_form is None:
            persistent_form = {}
        if form is None:
            form = {}
        super().__init__(DummyLexeme(form, persistent_form))

    def lexemes(self) -> Iterable[Lexeme]:
        for edge in self.edges:
            yield from edge.child.lexemes()


class SyntaxEdge:
    def __init__(self, parent: "SyntaxNode", child: "SyntaxNode", rel: Relation):
        self.parent = parent
        self.child = child
        self.rel = rel

    def agree(self):
        modifier = RelationDef.get(self.rel).modifier(
            self.parent.lexeme, self.child.lexeme
        )
        if modifier:
            self.child.lexeme.form = modifier
        self.child.agree()
        governing = RelationDef.get(self.rel).governing(
            self.child.lexeme, self.parent.lexeme
        )
        if governing:
            self.parent.lexeme.form = governing

    def __repr__(self):  # pragma: no cover
        return f"{self.rel} -> {repr(self.child)}"



import pytest
from .decimals import inflect
from .errors import RangeError

# fmt: off
simple_map = {
    "nom": ("case", "nominative"),
    "gen": ("case", "genitive"),
    "dat": ("case", "dative"),
    "acc": ("case", "accusative"),
    "inst": ("case", "instrumental"),
    "loc": ("case", "locative"),
    "voc": ("case", "vocative"),
    "m": ("gender", "masculine"),
    "f": ("gender", "feminine"),
    "n": ("gender", "neuter"),
    "sg": ("number", "singular"),
    "pl": ("number", "plural"),
    "card": ("type", "cardinal"),
    "ord": ("type", "ordinal"),
    "inan": ("animacy", "inanimate"),
    "anim": ("animacy", "animate"),
}


def fm(s):
    return {simple_map[g][0]: simple_map[g][1] for g in s.split(";")}


def test_gov_agreement():
    assert inflect("", "1", fm("m")) == "одна́ деся́та"
    assert inflect("", "2", fm("m")) == "дві́ деся́тих"
    assert inflect("", "3", fm("m")) == "три́ деся́тих"
    assert inflect("", "5", fm("m")) == "пʼя́ть деся́тих"
    assert inflect("", "0", fm("m")) == "ну́ль деся́тих"

def test_range_fallback():
    assert inflect("", "99999999999999999999999999", fm("m")) == "девʼяно́сто де́вʼять септильйо́нів девʼятсо́т девʼяно́сто де́вʼять секстильйо́нів девʼятсо́т девʼяно́сто де́вʼять квінтильйо́нів девʼятсо́т девʼяно́сто де́вʼять квадрильйо́нів девʼятсо́т девʼяно́сто де́вʼять трильйо́нів девʼятсо́т девʼяно́сто де́вʼять мілья́рдів девʼятсо́т девʼяно́сто де́вʼять мільйо́нів девʼятсо́т девʼяно́сто де́вʼять ти́сяч девʼятсо́т девʼяно́сто де́вʼять стасептильйо́нних"
    assert inflect("", "00000000000000000000000001", fm("m")) == "одна́ стасептильйо́нна"
    assert inflect("", "100000000000000000000000000", fm("m")) == "сто́ септильйо́нів оди́н ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль нульови́х"
    assert inflect("", "000000000000000000000000001", fm("m")) == "одна́ оди́н ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль нульова́"
    assert inflect("", "1000000000000000000000000000", fm("m")) == "оди́н ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль оди́н ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль нульова́"
    # assert inflect("", "1000000000000000000000000001", fm("m")) == "оди́н ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль оди́н"
    # assert inflect("", "1234567890123456789012345678", fm("m")) == "оди́н два́ три́ чоти́ри пʼя́ть ші́сть сі́м ві́сім де́вʼять ну́ль оди́н два́ три́ чоти́ри пʼя́ть ші́сть сі́м ві́сім де́вʼять ну́ль оди́н два́ три́ чоти́ри пʼя́ть ші́сть сі́м ві́сім"
    # assert inflect("", "999999999999999999999999999", fm("f")) == "девʼятсо́т девʼяно́сто де́вʼять септильйо́нів девʼятсо́т девʼяно́сто де́вʼять секстильйо́нів девʼятсо́т девʼяно́сто де́вʼять квінтильйо́нів девʼятсо́т девʼяно́сто де́вʼять квадрильйо́нів девʼятсо́т девʼяно́сто де́вʼять трильйо́нів девʼятсо́т девʼяно́сто де́вʼять мілья́рдів девʼятсо́т девʼяно́сто де́вʼять мільйо́нів девʼятсо́т девʼяно́сто де́вʼять ти́сяч девʼятсо́т девʼяно́сто де́вʼять"
    # assert inflect("", "1000000000000000000000000000", fm("f")) == "оди́н ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль"
    # assert inflect("", "1000000000000000000000000001", fm("f")) == "оди́н ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль одна́"
    # assert inflect("", "1234567890123456789012345678", fm("f")) == "оди́н два́ три́ чоти́ри пʼя́ть ші́сть сі́м ві́сім де́вʼять ну́ль оди́н два́ три́ чоти́ри пʼя́ть ші́сть сі́м ві́сім де́вʼять ну́ль оди́н два́ три́ чоти́ри пʼя́ть ші́сть сі́м ві́сім"
    # assert inflect("", "1000000000000000000000000000", fm("gen;f")) == "оди́н ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль нуля́"
    # assert inflect("", "1000000000000000000000000001", fm("gen;f")) == "оди́н ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль ну́ль одніє́ї"


def test_strict_range():
    assert inflect("", "99999999999999999999999999", fm("m"), strict_range=True)
    with pytest.raises(RangeError):
        assert inflect("", "100000000000000000000000000", fm("m"), strict_range=True)
